/*
	Copyright (C) 2020-2021 Reiko Asakura. All Rights Reserved.

	Music Premium
*/

#include <string.h>
#include <audioout.h>
#include <kernel/modulemgr.h>
#include <kernel/constant.h>
#include <kernel/sysmem.h>
#include <libdbg.h>
#include <sblacmgr.h>
#include <taihen.h>

#include "module.h"
#include "opcode.h"
#include "patch.h"

#define GLZ(x) do {\
	if ((x) < 0) { goto fail; }\
} while (0)

#define STREQ(s1, s2) (strncmp(s1, s2, sizeof(s1)) == 0)

SceUID bgm_port_priority_patch_id = -1;
SceUID sceAudioOutOpenPort_hook_id = -1;
tai_hook_ref_t sceAudioOutOpenPort_hook_ref;
SceUID sceAppMgrAcquireBgmPort_impl_hook_id = -1;
tai_hook_ref_t sceAppMgrAcquireBgmPort_impl_hook_ref;

static int (*sceAppMgrAcquireBgmPortWithPriority)(int priority);
static int (*sceAppMgrReleaseBgmPort)(void);

static int get_addr(uint16_t **inject_addr, uint16_t **hook_addr) {
	// get sceAppMgrAcquireBgmPort addr
	uint16_t *sceAppMgrAcquireBgmPort;
	if (GET_EXPORT("SceAppMgr", 0x8AF17416, 0xAFCEAB96, (void *)&sceAppMgrAcquireBgmPort) < 0) {
		return -1;
	}

	// clear Thumb bit
	sceAppMgrAcquireBgmPort = (uint16_t *)((uintptr_t)sceAppMgrAcquireBgmPort & ~0x1);

	// get sceAppMgrAcquireBgmPort impl func addr
	if (get_addr_bl(sceAppMgrAcquireBgmPort + 1, hook_addr) < 0) {
		return -1;
	}

	// get inject addr
	*inject_addr = *hook_addr + 0xE2/2;
	if ((*inject_addr)[0] != 0xF44F || (*inject_addr)[1] != 0x7380) {
		return -1;
	}

	// set Thumb bit
	*hook_addr = (uint16_t *)((uintptr_t)*hook_addr | 0x1);

	return 0;
}

static int sceAudioOutOpenPort_hook(int portType, int len, int freq, int param) {
	if (portType == SCE_AUDIO_OUT_PORT_TYPE_BGM && sceSblACMgrIsPspEmu(SCE_KERNEL_PROCESS_ID_SELF)) {
		tai_module_info_t minfo;
		minfo.size = sizeof(minfo);

		// redirect to voice port if not Adrenaline
		if (GET_MODULE(SCE_KERNEL_PROCESS_ID_SELF, "AdrenalineUser", &minfo) < 0) {
			portType = SCE_AUDIO_OUT_PORT_TYPE_VOICE;
		}
	}

	return HOOK_NEXT(sceAudioOutOpenPort, portType, len, freq, param);
}

static int sceAppMgrAcquireBgmPort_impl_hook(void) {
	int ret = HOOK_NEXT(sceAppMgrAcquireBgmPort_impl);

	if (sceSblACMgrIsShell(SCE_KERNEL_PROCESS_ID_SELF)) {
		char tid[0x10];
		sceKernelGetProcessTitleId(SCE_KERNEL_PROCESS_ID_SELF, tid, sizeof(tid));

		if (STREQ(tid, "VITASHELL") || STREQ(tid, "ELEVENMPV")) {
			if (ret == SCE_OK) {
				sceAppMgrReleaseBgmPort();
			}
			ret = sceAppMgrAcquireBgmPortWithPriority(0x82);
			SCE_DBG_LOG_INFO("BGM port acquired for music player %s", tid);
		}
	}

	return ret;
}

static void cleanup(void) {
	UNINJECT(bgm_port_priority);
	UNHOOK(sceAudioOutOpenPort);
	UNHOOK(sceAppMgrAcquireBgmPort_impl);
}

int module_start(SceSize argc, const void *argv) { (void)argc; (void)argv;

	uint16_t *inject_addr = NULL;
	uint16_t *hook_addr = NULL;

	// get SceAppMgr exports
	GLZ(GET_EXPORT("SceAppMgr", 0x8AF17416, 0xAAED7419, (void *)&sceAppMgrAcquireBgmPortWithPriority));
	GLZ(GET_EXPORT("SceAppMgr", 0x8AF17416, 0xF3717E37, (void *)&sceAppMgrReleaseBgmPort));

	// firmware independent way of getting absolute address
	GLZ(get_addr(&inject_addr, &hook_addr));

	// mov.w r3, #0x78
	GLZ(INJECT_ABS(inject_addr, "\x4f\xf0\x78\x03", 4, bgm_port_priority));

	GLZ(HOOK_EXPORT("SceAudio", 0x438BB957, 0x5BC341E4, sceAudioOutOpenPort));
	GLZ(HOOK_ABS(hook_addr, sceAppMgrAcquireBgmPort_impl));

	return SCE_KERNEL_START_SUCCESS;

fail:
	cleanup();
	return SCE_KERNEL_START_FAILED;
}

int module_stop(SceSize argc, const void *argv) { (void)argc; (void)argv;
	cleanup();
	return SCE_KERNEL_STOP_SUCCESS;
}
